from contextlib import asynccontextmanager
from typing import AsyncContextManager

from nonebot import get_driver
from nonebot import logger, Driver
from sqlalchemy.ext.asyncio import AsyncSession, AsyncEngine, create_async_engine
from sqlalchemy.orm import registry
from sqlalchemy.orm import sessionmaker

from ..config import conf


class DataSourceNotReadyError(RuntimeError):
    pass


class DataSource:
    def __init__(self, driver: Driver, url: str, **kwargs):
        self._engine = None
        self._sessionmaker = None

        self._registry = registry()

        # 仅当TRACE模式时回显sql语句
        kwargs.setdefault("echo", driver.config.log_level == 'TRACE')
        kwargs.setdefault("future", True)

        @driver.on_startup
        async def on_startup():
            self._engine = create_async_engine(url, **kwargs)

            async with self._engine.begin() as conn:
                await conn.run_sync(self._registry.metadata.create_all)

            # expire_on_commit=False will prevent attributes from being expired
            # after commit.
            self._sessionmaker = sessionmaker(
                self._engine, expire_on_commit=False, class_=AsyncSession
            )
            logger.success("data source initialized")

        @driver.on_shutdown
        async def on_shutdown():
            await self._engine.dispose()

            self._engine = None
            self._sessionmaker = None

            logger.success("data source disposed")

    @property
    def engine(self) -> AsyncEngine:
        if self._engine is None:
            raise DataSourceNotReadyError()
        return self._engine

    @property
    def registry(self) -> registry:
        return self._registry

    @asynccontextmanager
    async def start_session(self) -> AsyncContextManager[AsyncSession]:
        if self._sessionmaker is None:
            raise DataSourceNotReadyError()

        async with self._sessionmaker() as session:
            yield session


data_source = DataSource(get_driver(), conf.access_control_database_conn_url)

__all__ = ("data_source", "DataSource", "DataSourceNotReadyError")



from . import access_control
from . import config
from . import errors
from . import event_bus
from . import matchers
from . import models
from . import service
from . import subject
from . import utils



from nonebot.rule import ArgumentParser

parser = ArgumentParser(prog="rbac")
subparsers = parser.add_subparsers(help="子命令", required=True, dest="subcommand")

# ==== subject ====
subject_parser = subparsers.add_parser("subject", help="主体管理")
subject_parser.add_argument("subject", help="主体")
subject_subparsers = subject_parser.add_subparsers(help="操作", required=True, dest="action")

#     ==== subject <xxx> ls ====
subject_ls_parser = subject_subparsers.add_parser("ls", help="列出")
subject_ls_subparsers = subject_ls_parser.add_subparsers(help="对象", required=True, dest="target")

#         ==== subject <xxx> ls service ====
subject_ls_service_parser = subject_ls_subparsers.add_parser("service", help="列出主体可用服务")

#     ==== subject <xxx> remove <xxx> ====
subject_remove_parser = subject_subparsers.add_parser("remove", help="为主体移除服务的权限设置")
subject_remove_subparsers = subject_remove_parser.add_subparsers(help="对象", dest="target")

#         ==== subject <xxx> remove service <xxx> ====
subject_remove_service_parser = subject_remove_subparsers.add_parser("service", help="列出主体可用服务")
subject_remove_service_parser.add_argument("service", help="服务")

#     ==== subject <xxx> allow <xxx> ====
subject_allow_parser = subject_subparsers.add_parser("allow", help="为主体启用服务")
subject_allow_subparsers = subject_allow_parser.add_subparsers(help="对象", dest="target")

#         ==== subject <xxx> allow service <xxx> ====
subject_allow_service_parser = subject_allow_subparsers.add_parser("service", help="为主体启用服务")
subject_allow_service_parser.add_argument("service", help="服务")

#     ==== subject <xxx> deny <xxx> ====
subject_deny_parser = subject_subparsers.add_parser("deny", help="为主体禁用服务")
subject_deny_subparsers = subject_deny_parser.add_subparsers(help="对象", dest="target")

#         ==== subject <xxx> deny service <xxx> ====
subject_deny_service_parser = subject_deny_subparsers.add_parser("service", help="为主体禁用服务")
subject_deny_service_parser.add_argument("service", help="服务")

# ==== service ====
service_parser = subparsers.add_parser("service", help="主体管理")
service_parser.add_argument("service", help="主体")
service_subparsers = service_parser.add_subparsers(help="操作", required=True, dest="action")

#     ==== service <xxx> ls ====
service_ls_parser = service_subparsers.add_parser("ls", help="列出")
service_ls_subparsers = service_ls_parser.add_subparsers(help="对象", dest="target")

#         ==== service <xxx> ls subject ====
service_ls_subject_parser = service_ls_subparsers.add_parser("subject", help="列出服务的主体权限设置")



from io import StringIO
from typing import Protocol, Optional, Collection, TypeVar, Callable

from typing_extensions import Self


class TreeNode(Protocol):
    parent: Optional[Self]
    children: Collection[Self]
    name: str


T = TypeVar("T")


def get_tree_summary(root: T,
                     children: Callable[[T], Collection[T]],
                     value: Callable[[T], str]) -> str:
    with StringIO() as sio:
        def walk(node: T, prefix: str = ""):
            node_children = children(node)
            for i, child in enumerate(node_children):
                if i == len(node_children) - 1:
                    sio.write(prefix + "└── " + value(child) + '\n')
                    walk(child, prefix + "    ")
                else:
                    sio.write(prefix + "├── " + value(child) + '\n')
                    walk(child, prefix + "│   ")

        sio.write(value(root) + '\n')
        walk(root)
        return sio.getvalue().strip()



from asyncio import gather
from collections import defaultdict
from enum import Enum
from inspect import isawaitable
from typing import Dict, List, Callable, Awaitable, Any, Tuple, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    pass


class EventType(str, Enum):
    service_allow_permission = "service_allow_permission"
    service_deny_permission = "service_deny_permission"
    service_remove_permission = "service_remove_permission"


T_Kwargs = Dict[str, Any]
T_Filter = Callable[[T_Kwargs], bool]
T_Listener = Callable[[...], Awaitable[None]]

_listeners: Dict[EventType, List[Tuple[T_Filter, T_Listener]]] = defaultdict(list)


async def fire_event(event_type: EventType, kwargs: T_Kwargs):
    coros = []

    for filter_func, func in _listeners[event_type]:
        if filter_func(kwargs):
            coro = func(**kwargs)
            if isawaitable(coro):
                coros.append(coro)

    await gather(*coros)


def on_event(event_type: EventType, filter_func: T_Filter, func: Optional[T_Listener] = None):
    def decorator(func):
        _listeners[event_type].append((filter_func, func))
        return func

    if func is None:
        return decorator
    else:
        return decorator(func)



from typing import Type, Dict

import nonebot
from nonebot import get_driver, logger
from nonebot.adapters.onebot.v11 import Bot
from nonebot.exception import IgnoredException
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.message import run_preprocessor

from .config import conf
from .service import Service, get_plugin_service

_matcher_service_mapping: Dict[Type[Matcher], Service] = {}


def patch_matcher(matcher: Type[Matcher], service: Service) -> Type[Matcher]:
    _matcher_service_mapping[matcher] = service
    logger.debug(f"patched {matcher}  (with service {service.qualified_name})")
    return matcher


@run_preprocessor
async def check(matcher: Matcher, bot: Bot, event: Event):
    service = _matcher_service_mapping.get(type(matcher), None)
    if service is None:
        return

    allow = await service(bot, event)

    if not allow:
        raise IgnoredException("permission denied by nonebot_plugin_access_control")


if conf.access_control_auto_patch_enabled:
    @get_driver().on_startup
    def _do_patch():
        for plugin in nonebot.get_loaded_plugins():
            if plugin.name == 'nonebot_plugin_access_control' or plugin.name in conf.access_control_auto_patch_ignore:
                continue

            service = get_plugin_service(plugin.name)
            if service.auto_created:
                for matcher in plugin.matcher:
                    patch_matcher(matcher, service)



from urllib.parse import urlparse


def detect_sql_dialect(conn_url: str) -> str:
    url = urlparse(conn_url)
    if '+' in url.scheme:
        return url.scheme.split('+')[0]
    else:
        return url.scheme



from typing import Callable, Dict, List

from nonebot import Bot, logger
from nonebot.internal.adapter import Event

T_SubjectExtractor = Callable[[Bot, Event], List[str]]

_subject_extractors: Dict[str, T_SubjectExtractor] = {}


def register_subject_extractor(adapter: str, subject_extractor: T_SubjectExtractor):
    _subject_extractors[adapter] = subject_extractor


def extract_subjects(bot: Bot, event: Event) -> List[str]:
    adapter = bot.adapter.get_name()
    subject_extractor = _subject_extractors.get(adapter, None)
    if subject_extractor is None:
        logger.warning(f"no subject extractor found for adapter {adapter}")
        return [f"{adapter}:unknown"]
    return subject_extractor(bot, event)


__all__ = ("register_subject_extractor", "extract_subjects")



from functools import wraps

from nonebot import logger
from nonebot.exception import MatcherException
from nonebot.internal.matcher import current_matcher

from ..errors import RbacError


def handle_error(silently: bool = False):
    def decorator(func):
        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except MatcherException as e:
                raise e
            except RbacError as e:
                logger.exception(e)
                if not silently:
                    matcher = current_matcher.get()
                    await matcher.finish(f"{str(e)}")
            except Exception as e:
                logger.exception(e)
                if not silently:
                    matcher = current_matcher.get()
                    await matcher.finish(f"内部错误：{type(e)}{str(e)}")

        return wrapped_func

    return decorator



from sqlalchemy import Column, String, Boolean

from .data_source import data_source


@data_source.registry.mapped
class PermissionOrm:
    __tablename__ = 'permissions'

    subject: str = Column(String, nullable=False, primary_key=True)
    service: str = Column(String, nullable=False, primary_key=True)
    allow: bool = Column(Boolean, nullable=False)



from argparse import Namespace
from io import StringIO
from typing import cast

from nonebot import on_shell_command
from nonebot.exception import ParserExit
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.permission import SUPERUSER
from nonebot.typing import T_State

from .handle_error import handle_error
from .parser import parser
from ..service import get_services_by_subject, Service, get_service_by_qualified_name
from ..utils.tree import get_tree_summary

cmd = on_shell_command("ac", parser=parser, permission=SUPERUSER)


@cmd.handle()
@handle_error()
async def _(matcher: Matcher, event: Event, state: T_State):
    args = state["_args"]
    if isinstance(args, ParserExit):
        await matcher.finish(args.message)

    args = cast(Namespace, args)

    if args.subcommand == 'subject':
        if args.action == 'ls':
            if args.target == 'service':
                await handle_subject_ls_service(matcher, args.subject)
        elif args.action == 'allow':
            if args.target == 'service':
                await handle_subject_allow_service(matcher, args.subject, args.service)
        elif args.action == 'deny':
            if args.target == 'service':
                await handle_subject_deny_service(matcher, args.subject, args.service)
        elif args.action == 'remove':
            if args.target == 'service':
                await handle_subject_remove_service(matcher, args.subject, args.service)
    elif args.subcommand == 'service':
        if args.action == 'ls':
            if args.target is None:
                await handle_service_ls_subservice(matcher, args.service)
            elif args.target == 'subject':
                await handle_service_ls_subject(matcher, args.service)


async def handle_subject_ls_service(matcher: Matcher, subject: str):
    services = [x async for x in get_services_by_subject(subject)]
    if len(services) != 0:
        # 按照先allow再deny排序
        services = sorted(services, key=lambda x: x[1], reverse=True)
        msg = '\n'.join(map(lambda x: x[0].qualified_name + (' allow' if x[1] else ' deny'), services))
    else:
        msg = "empty"
    await matcher.send(msg)


async def _get_service(matcher: Matcher, service_name: str) -> Service:
    service_name = get_service_by_qualified_name(service_name)

    if service_name is None:
        await matcher.finish("service not found")
    else:
        return service_name


async def handle_subject_allow_service(matcher: Matcher, subject: str, service: str):
    service = await _get_service(matcher, service)
    await service.set_permission(subject, True)
    await matcher.send("ok")


async def handle_subject_deny_service(matcher: Matcher, subject: str, service: str):
    service = await _get_service(matcher, service)
    await service.set_permission(subject, False)
    await matcher.send("ok")


async def handle_subject_remove_service(matcher: Matcher, subject: str, service: str):
    service = await _get_service(matcher, service)
    await service.remove_permission(subject)
    await matcher.send("ok")


async def handle_service_ls_subject(matcher: Matcher, service_name: str):
    permissions = {}
    service = await _get_service(matcher, service_name)

    while service is not None:
        async for subject, allow in service.get_permissions():
            permissions.setdefault(subject, (allow, service))
        service = service.parent

    if len(permissions) != 0:
        # 按照先allow再deny排序
        permissions = [(*permissions[k], k) for k in permissions]
        permissions = sorted(permissions, reverse=True)
        with StringIO() as sio:
            for allow, service, subject in permissions:
                sio.write(subject)
                if allow:
                    sio.write(" allow")
                else:
                    sio.write(" deny")

                if service.qualified_name != service_name:
                    sio.write(f" (inherited from {service.qualified_name})")
                sio.write('\n')
            msg = sio.getvalue().strip()
    else:
        msg = "empty"
    await matcher.send(msg)


async def handle_service_ls_subservice(matcher: Matcher, service_name: str):
    service = await _get_service(matcher, service_name)
    summary = get_tree_summary(service, lambda x: x.children, lambda x: x.name)
    await matcher.send(summary)



from typing import Literal, List

from nonebot import get_driver
from pydantic import BaseSettings, Field


class Config(BaseSettings):
    access_control_database_conn_url: str = "sqlite+aiosqlite:///access_control.db"
    access_control_default_permission: Literal['allow', 'deny'] = 'allow'
    access_control_auto_patch_enabled: bool = True
    access_control_auto_patch_ignore: List[str] = Field(default_factory=list)

    class Config:
        extra = "ignore"


conf = Config(**get_driver().config.dict())



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nonebot_plugin_access_control',
 'nonebot_plugin_access_control.matchers',
 'nonebot_plugin_access_control.models',
 'nonebot_plugin_access_control.subject',
 'nonebot_plugin_access_control.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiosqlite>=0.17.0,<0.18.0',
 'nonebot2>=2.0.0rc1,<3.0.0',
 'sqlalchemy[asyncio]>=1.4.42,<2.0.0']

setup_kwargs = {
    'name': 'nonebot-plugin-access-control',
    'version': '0.0.2',
    'description': '',
    'long_description': '<!-- markdownlint-disable MD033 MD036 MD041 -->\n\n<p align="center">\n  <a href="https://v2.nonebot.dev/"><img src="https://v2.nonebot.dev/logo.png" width="200" height="200" alt="nonebot"></a>\n</p>\n\n<div align="center">\n\nnonebot-plugin-access-control\n============\n\n_✨ Nonebot 权限控制 ✨_\n\n</div>\n\n\n<p align="center">\n  <a href="https://raw.githubusercontent.com/ssttkkl/nonebot-plugin-access-control/master/LICENSE">\n    <img src="https://img.shields.io/github/license/ssttkkl/nonebot-plugin-access-control.svg" alt="license">\n  </a>\n  <a href="https://pypi.python.org/pypi/nonebot-plugin-access-control">\n    <img src="https://img.shields.io/pypi/v/nonebot-plugin-access-control.svg" alt="pypi">\n  </a>\n  <img src="https://img.shields.io/badge/python-3.9+-blue.svg" alt="python">\n</p>\n\n## 特点\n\n- [x] 支持**功能级别**的细粒度权限控制\n- [x] 对未适配插件也支持**插件级别**的权限控制\n- [ ] 支持对权限开关等事件进行监听\n\n## 使用\n\n### 主体\n\n#### 概念\n\n让我们从一个例子引入：QQ上群组G的用户U发送了一条消息，该用户同时具有“用户U”、“群组G成员”、“QQ用户”、“Bot用户”这几个身份。同时QQ上群组G的用户V也发送了一条消息，该用户该用户同时具有“用户V”、“群组G成员”、“QQ用户”、“Bot用户”这几个身份。\n\n如果我们希望对用户进行权限控制，我们可以直接针对“用户U”、“用户V”这个级别分别配置权限。但我们希望对群组进行权限控制时，对群组内的每一个用户都分别配置权限，就有点力不从心了。我们希望能够直接针对“群组G”这个级别进行配置，而无需关心群组内都有什么成员。\n\n我们定义**主体（Subject）为用户所具有的身份，也是设置权限的基本单位。**\n\n一个用户通常拥有多个主体。回到上面的例子，第一位用户拥有“用户U”、“群组G成员”、“QQ用户”、“Bot用户”这四个主体；第二位用户拥有“用户V”、“群组G成员”、“QQ用户”、“Bot用户”这四个主体。用户拥有的所有主体，按照规模从小到大排序，呈现如下图的逐层包含关系：\n\n![](docs/img/1.svg)\n\n当设置权限时，我们直接针对一个主体进行设置。当鉴权时，我们对用户的所有主体按规模从小到大的顺序（下文将此顺序称为优先级顺序），逐一检查是否设置了权限。一旦检查到某个主体设置了权限，就以该主体设置的权限作为该用户的权限。\n\n回到上面的例子，假设我们对主体”群组G“禁用服务，但是对主体”用户V“启用服务。则在群组G内的用户U内将无法使用服务，但是则在群组G内的用户V可以使用。\n\n#### 应用\n\n在实际应用中，我们用一个字符串表示主体。并且我们约定，`all`表示所有用户，`<协议名>`表示所有此协议的用户，除此之外的所有主体均以`<协议名>:`开头。\n\n在OneBot协议中，每个用户所拥有的主体如下表所定义：\n\n| 主体                 | 含义       | 示例               | 必定存在             |\n|--------------------|----------|------------------|------------------|\n| onebot:<user_id>   | 用户       | onebot:12345678  | 是                |\n| onebot:g<group_id> | 群组       | onebot:g87654321 | 仅当消息来自群组或临时会话时存在 |\n| onebot             | OneBot用户 | onebot           | 是                |\n| all                | 所有用户     | all              | 是                |\n\n目前仅实现了OneBot协议。如果你能帮助我们进行其他协议适配，欢迎提交PR。\n\n### 服务\n\n**服务（Service）为一组能够进行权限控制的功能的集合。** 服务可以拥有子服务，通过树形结构组织服务，统一管理权限。\n\n整个插件就是一个服务（PluginService）。当插件未进行适配时，该插件只具有一个PluginService。\n\n若需要对插件进行适配，则需要从PluginService创建SubService，为插件的Matcher等功能入口应用SubService。（参考下文插件适配章节）\n\n### 指令\n\n进行权限开关的指令为`/ac`，仅超级用户可用。（通过在配置文件中设置`SUPERUSERS`变量可设置超级用户）\n\n- `/ac subject <主体> allow service <服务>`：为主体启用服务\n- `/ac subject <主体> deny service <服务>`：为主体禁用服务\n- `/ac subject <主体> remove service <服务>`：为主体删除服务权限配置\n- `/ac subject <主体> ls service`：列出主体已配置的服务权限\n- `/ac service <服务> ls`：列出服务的子服务层级\n- `/ac service <服务> ls subject`：列出服务已配置的主体权限\n\n其中`<服务>`的格式如下：\n\n- `<插件名>`：对整个插件进行开关\n- `<插件名>.<子服务名>.<子服务名>.....<子服务名>`：对插件内的某个子服务进行开关（需参照下文对插件进行配置）\n\n## 插件适配\n\n完整代码：[src/nonebot_plugin_ac_demo](src/nonebot_plugin_ac_demo)\n\n1. 创建一个名为nonebot_plugin_ac_demo的插件\n\n2. 通过create_plugin_service函数创建一个PluginServic实例（注意参数必须为插件包名）\n\n```python\nfrom nonebot import require\n\nrequire("nonebot_plugin_ac_demo")\n\nfrom nonebot_plugin_access_control.service import create_plugin_service\n\nplugin_service = create_plugin_service("nonebot_plugin_ac_demo")\n```\n\n3. 通过PluginService.create_subservice创建SubService实例。调用`SubService.patch_matcher()`应用至Matcher\n\n```python\ngroup1 = plugin_service.create_subservice("group1")\n\na_matcher = on_command(\'a\')\na_service = group1.create_subservice(\'a\')\na_service.patch_matcher(a_matcher)\n\n@a_matcher.handle()\nasync def _(matcher: Matcher):\n    await matcher.send("a")\n\n    \nb_matcher = on_command(\'b\')\nb_service = group1.create_subservice(\'b\')\nb_service.patch_matcher(b_matcher)\n\n@b_matcher.handle()\nasync def _(matcher: Matcher):\n    await matcher.send("b")\n\n    \nc_matcher = on_command(\'c\')\nc_service = plugin_service.create_subservice(\'c\')\nc_service.patch_matcher(c_matcher)\n\n@c_matcher.handle()\nasync def _(matcher: Matcher):\n    await matcher.send("c")\n```\n\n插件服务的结构如下所示：\n\n![](docs/img/2.svg)\n\n4. 通过指令配置服务权限\n\n执行下面的指令后，所有用户将无法调用指令`/a`与`/b`\n\n```\n/ac subject all deny service nonebot_plugin_ac_demo.group1\n```\n\n执行下面的指令后，用户12345678将无法调用指令`/a`\n\n```\n/ac subject onebot:12345678 deny service nonebot_plugin_ac_demo.group1.a\n```\n\n执行下面的指令后，群组87654321的所有用户将无法调用除`/c`以外的任何指令\n\n```\n/ac subject onebot:g87654321 deny service nonebot_plugin_ac_demo\n/ac subject onebot:g87654321 allow service nonebot_plugin_ac_demo.c\n```',
    'author': 'ssttkkl',
    'author_email': 'huang.wen.long@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ssttkkl/nonebot-plugin-access-control',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import re
from abc import ABC, abstractmethod
from typing import Optional, Dict, AsyncGenerator, Tuple, overload, Literal, Callable, Awaitable, Type, Collection

import conf as conf
import nonebot
from nonebot import Bot, logger
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from sqlalchemy import select, delete
from sqlalchemy.ext.asyncio import AsyncSession

from .config import conf
from .errors import RbacError
from .event_bus import on_event, EventType, T_Listener, fire_event
from .models import data_source, PermissionOrm
from .subject import extract_subjects


def _validate_name(name: str) -> bool:
    match_result = re.match(r"[_a-zA-Z]\w*", name)
    return match_result is not None


class Service(ABC):
    def __init__(self):
        self._subservices: Dict[str, Service] = {}

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    def qualified_name(self) -> str:
        if self.parent is None:
            return self.name
        else:
            return self.parent.qualified_name + "." + self.name

    @property
    @abstractmethod
    def plugin_service(self) -> "PluginService":
        raise NotImplementedError()

    @property
    @abstractmethod
    def parent(self) -> "Optional[Service]":
        raise NotImplementedError()

    @property
    def children(self) -> "Collection[Service]":
        return self._subservices.values()

    def create_subservice(self, name: str) -> "Service":
        if not _validate_name(name):
            raise RbacError(f'invalid name: {name}')

        service = SubService(name, self.plugin_service, self)
        self._subservices[name] = service

        logger.trace(f"created subservice {service.qualified_name}  (parent: {self.qualified_name})")
        return self._subservices[name]

    def get_subservice(self, name: str) -> "Optional[Service]":
        return self._subservices.get(name, None)

    def travel(self):
        sta = [self]
        while len(sta) != 0:
            top, sta = sta[-1], sta[:-1]
            yield top
            sta.extend(top._subservices.values())

    def find(self, name: str) -> "Optional[Service]":
        for s in self.travel():
            if s.name == name:
                return s
        return None

    def patch_matcher(self, matcher: Type[Matcher]) -> Type[Matcher]:
        from .access_control import patch_matcher
        return patch_matcher(matcher, self)

    def on_allow_permission(self, func: Optional[T_Listener] = None):
        return on_event(EventType.service_allow_permission,
                        lambda kwargs: kwargs["service"] == self,
                        func)

    def on_deny_permission(self, func: Optional[T_Listener] = None):
        return on_event(EventType.service_deny_permission,
                        lambda kwargs: kwargs["service"] == self,
                        func)

    def on_remove_permission(self, func: Optional[T_Listener] = None):
        return on_event(EventType.service_remove_permission,
                        lambda kwargs: kwargs["service"] == self,
                        func)

    async def _get_permission(self, node_permission_getter: Callable[["Service", AsyncSession],
                                                                     Awaitable[Optional[bool]]]) -> Optional[bool]:
        async with data_source.start_session() as session:
            v = self
            allow = None

            while v is not None:
                p = await node_permission_getter(v, session)
                if p is not None:
                    allow = p
                    break
                else:
                    v = v.parent

            return allow

    @overload
    async def get_permission(self, subject: str) -> bool:
        ...

    @overload
    async def get_permission(self, subject: str, with_default: Literal[True]) -> bool:
        ...

    @overload
    async def get_permission(self, subject: str, with_default: Literal[False]) -> Optional[bool]:
        ...

    async def get_permission(self, subject: str, with_default: bool = True) -> Optional[bool]:
        async def _get_permission(node: "Service", session: AsyncSession) -> Optional[bool]:
            stmt = (select(PermissionOrm)
                    .where(PermissionOrm.service == node.qualified_name,
                           PermissionOrm.subject == subject))
            p = (await session.execute(stmt)).scalar_one_or_none()
            if p is not None:
                return p.allow
            else:
                return None

        allow = await self._get_permission(_get_permission)
        if allow is None and with_default:
            allow = conf.access_control_default_permission == 'allow'
        return allow

    async def get_permissions(self) -> AsyncGenerator[Tuple[str, bool], None]:
        async with data_source.start_session() as session:
            stmt = (select(PermissionOrm)
                    .where(PermissionOrm.service == self.qualified_name))
            async for p in await session.stream_scalars(stmt):
                yield p.subject, p.allow

    async def _fire_service_set_permission(self, subject: str, allow: bool):
        if allow:
            event_type = EventType.service_allow_permission
        else:
            event_type = EventType.service_deny_permission

        for node in self.travel():
            await fire_event(event_type, {
                "service": node,
                "subject": subject
            })

    async def set_permission(self, subject: str, allow: bool):
        async with data_source.start_session() as session:
            stmt = (select(PermissionOrm)
                    .where(PermissionOrm.service == self.qualified_name,
                           PermissionOrm.subject == subject))
            p = (await session.execute(stmt)).scalar_one_or_none()
            if p is None:
                p = PermissionOrm(service=self.qualified_name,
                                  subject=subject,
                                  allow=allow)
                session.add(p)
                ok = True
            else:
                ok = p.allow != allow
                p.allow = allow

            if ok:
                await session.commit()
                await self._fire_service_set_permission(subject, allow)

            return ok

    async def _fire_service_remove_permission(self, subject: str):
        for node in self.travel():
            await fire_event(EventType.service_remove_permission, {
                "service": node,
                "subject": subject
            })

        await self._fire_service_set_permission(subject, conf.access_control_default_permission == 'allow')

    async def remove_permission(self, subject: str) -> bool:
        async with data_source.start_session() as session:
            stmt = (delete(PermissionOrm)
                    .where(PermissionOrm.service == self.qualified_name,
                           PermissionOrm.subject == subject))
            row_count = (await session.execute(stmt)).rowcount
            ok = row_count == 1

            if ok:
                await session.commit()
                await self._fire_service_remove_permission(subject)

            return ok

    async def __call__(self, bot: Bot, event: Event) -> bool:
        subjects = extract_subjects(bot, event)
        for sub in subjects:
            p = await self.get_permission(sub, False)
            if p is not None:
                return p

        return conf.access_control_default_permission == 'allow'


class PluginService(Service):
    def __init__(self, name: str, auto_created: bool):
        super().__init__()
        self._name = name
        self._auto_created = auto_created

    @property
    def name(self) -> str:
        return self._name

    @property
    def plugin_service(self) -> "PluginService":
        return self

    @property
    def parent(self) -> "Optional[Service]":
        return None

    @property
    def auto_created(self) -> bool:
        return self._auto_created


class SubService(Service):
    def __init__(self, name: str, plugin_service: PluginService, parent: Optional[Service] = None):
        super().__init__()
        self._name = name
        self._plugin_service = plugin_service
        self._parent = parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def plugin_service(self) -> "PluginService":
        return self._plugin_service

    @property
    def parent(self) -> "Optional[Service]":
        return self._parent


_plugin_services: Dict[str, PluginService] = {}


def _create_plugin_service(plugin_name: str, auto_create: bool) -> PluginService:
    if plugin_name in _plugin_services:
        raise ValueError(f"{plugin_name} already created")

    service = PluginService(plugin_name, auto_create)
    _plugin_services[plugin_name] = service
    logger.trace(f"created plugin service {service.qualified_name}")
    return service


def create_plugin_service(plugin_name: str) -> PluginService:
    return _create_plugin_service(plugin_name, False)


def get_plugin_service(plugin_name: str, auto_create: bool = True) -> Optional[PluginService]:
    if plugin_name in _plugin_services:
        return _plugin_services[plugin_name]
    elif auto_create:
        plugin = nonebot.get_plugin(plugin_name)
        if plugin is not None:
            return _create_plugin_service(plugin_name, auto_create)

    return None


def get_service_by_qualified_name(qualified_name: str, auto_create_plugin_service: bool = True) -> Optional[Service]:
    seg = qualified_name.split('.')
    service = get_plugin_service(seg[0], auto_create_plugin_service)
    for i in range(1, len(seg)):
        if service is None:
            return None
        service = service.get_subservice(seg[i])
    return service


async def get_services_by_subject(subject: str) -> AsyncGenerator[Tuple[Service, bool], None]:
    async with data_source.start_session() as session:
        stmt = select(PermissionOrm).where(PermissionOrm.subject == subject)
        async for x in await session.stream_scalars(stmt):
            service = get_service_by_qualified_name(x.service)
            yield service, x.allow


__all__ = ("create_plugin_service", "get_plugin_service", "get_services_by_subject",
           "Service", "PluginService", "SubService")



from nonebot import logger

from .extractor import register_subject_extractor

try:
    from nonebot.adapters.kaiheila import Bot, Event
    from nonebot.adapters.kaiheila.adapter import Adapter


    def kaiheila_subject_extractor(bot: Bot, event: Event):
        li = []

        user_id = getattr(event, "user_id", None)
        if user_id is not None:
            li.append(f"kaiheila:{user_id}")

        channel_id = getattr(event, "channel_id", None)
        if channel_id is not None:
            li.append(f"kaiheila:c{channel_id}")

        if event.extra.guild_id is not None:
            li.append(f"kaiheila:g{event.extra.guild_id}")

        li.append("kaiheila")
        li.append("all")

        return li


    register_subject_extractor(Adapter.get_name(), kaiheila_subject_extractor)
    logger.trace(f"registered subject extractor for {Adapter.get_name()}")
except ImportError:
    pass



from nonebot import logger

from .extractor import register_subject_extractor

try:
    from nonebot.adapters.onebot.v11 import Bot, Event
    from nonebot.adapters.onebot.v11.adapter import Adapter


    def onebot_v11_subject_extractor(bot: Bot, event: Event):
        user_id = getattr(event, "user_id", None)
        group_id = getattr(event, "group_id", None)

        li = []

        if user_id is not None:
            li.append(f"onebot:{user_id}")

        if group_id is not None:
            li.append(f"onebot:g{group_id}")

        li.append("onebot")
        li.append("all")

        return li


    register_subject_extractor(Adapter.get_name(), onebot_v11_subject_extractor)
    logger.trace(f"registered subject extractor for {Adapter.get_name()}")
except ImportError:
    pass
